// Author: Sergey Smirnov

#include <avr/interrupt.h>
#include "structs.h"

const uint8_t RISING_EDGE = 1;
const uint8_t FALLING_EDGE = 0;

const uint8_t IDLE_PERIODS = 2;

inline void setEdge(uint8_t edge) {
  TCCR1B = 0x01 | (edge << 6); // no Noice canceler + ICE=RISING_EDGE/FALLING_EDGE + System timer
}

Result results[2];
Result *current = results;
Result *collected = results + 1;

short resultUpdated = 0;
short saySomething = 0;
short state = 0;
short idlePeriods = IDLE_PERIODS;

uint8_t maxOverflows = 8;
uint8_t sensorCount = 1;

inline void setupNextMux() {
  current->downs.overflows = 0;
  
  ++current->mux;
  if (current->mux >= sensorCount) {
    current->mux = 0;
  }
  ADMUX = current->mux;
  
  setEdge(FALLING_EDGE);
  idlePeriods = IDLE_PERIODS;
  state = 7;
}

inline void swapResults() {
  Result *tmp = current;
  current = collected;
  collected = tmp;
  current->mux = collected->mux;
}



ISR(TIMER1_CAPT_vect) {
  unsigned int currentTime = ICR1;
  
  switch (state) {
    case 7:
      if (--idlePeriods <= 0) {
        state = 0;
      }
      break;
    case 0:
      current->downs.startTime = currentTime;
      current->downs.overflows = 0;
      current->downs.tempOverflows = 0;
      setEdge(RISING_EDGE);
      state = 1;
      break;
    case 1:
      current->ups.startTime = currentTime;
      current->ups.overflows = 0;
      current->periods = 0;
      setEdge(FALLING_EDGE);
      state = 2;
      break;
    case 2:
      current->downs.endTime = currentTime;
      current->downs.overflows = current->downs.tempOverflows;
      setEdge(RISING_EDGE);
      state = 3;
      break;
    case 3:
      current->ups.endTime = currentTime;
      ++current->periods;
      
      if (current->downs.overflows >= maxOverflows) {
        swapResults();
        resultUpdated = 1;
        
        setupNextMux();
      } else {
        setEdge(FALLING_EDGE);
        state = 2;
      }
      break;
  }
}

ISR(TIMER1_OVF_vect) {
  ++current->downs.tempOverflows;
  ++current->ups.overflows;
  
  if (current->ups.overflows >= maxOverflows + 1) {
    saySomething = 1;
    setupNextMux();
  }
}

void setup() {
  Serial.begin(57600);
  
  // Setup comparator:
  ADCSRA &= ~0x80; // disable adc to use its multiplexer
  
  // enable multiplexer input for comparator (insted of AIN1 (negative))
  #ifdef __AVR_ATmega8__
    SFIOR |= 0x08;  // for ATmega8
  #else
    ADCSRB |= 0x40; // for ATmega168
  #endif
  
  ACSR = 0x44; // comparator: enable, use bandgap voltage on AIN0 (positive), connect to ICU
  ADMUX = 0x00; // select input in multiplexer, need only first three bits
  
  // unset Input Capture Flag
  #ifdef __AVR_ATmega8__
    TIFR &= ~(1 << 5); // for ATmega8
  #else
    TIFR1 &= ~(1 << 5); // for ATmega168
  #endif
  
  // Setup Timer/Counter1
  TCCR1A = 0; // Normal mode (without PWM, TOP=0xffff)
  setEdge(FALLING_EDGE); //TCCR1B = 0x01; // no Noice canceler + ICE=FALLING_EDGE + System timer
  
  // enable event + overflow interrupts
  #ifdef __AVR_ATmega8__
    TIMSK |= 0x24; // for ATmega8
  #else
    TIMSK1 |= (1 << 5) | (1 << 0); // for ATmega168
  #endif
}

inline void printWord(unsigned int word) {
    Serial.print((word >> 0) & 0x7F, BYTE);
    Serial.print((word >> 7) & 0x7F, BYTE);
}

inline unsigned long calculateTicks(Result *result) {
  unsigned long ticks = result->ups.overflows + result->downs.overflows;
  ticks <<= 16;
  ticks -= result->ups.startTime;
  ticks += result->ups.endTime;
  ticks -= result->downs.startTime;
  ticks += result->downs.endTime;
  return ticks;
}

inline void printResultCod(Result *result) {
  if (result == NULL) {
    Serial.print(0x80, BYTE);
    return;
  }
  
  Serial.print(result->mux & 0x7F, BYTE);
  printWord(result->periods << 1);
  
  unsigned long ticks = calculateTicks(result);
  
  printWord(ticks);
  printWord(ticks >> 14);
  
  Serial.print(0x80, BYTE);
}

inline void printResultText(Result *result) {
  if (result == NULL) {
    Serial.println("");
    return;
  }
  
  Serial.print((int)result->mux);
  Serial.print(' ');
  Serial.print(result->periods << 1);
  Serial.print(' ');
  Serial.println(calculateTicks(result));
}

inline void printResult(Result *result) {
  //printResultText(result);
  printResultCod(result);
}


void loop () {
  if (resultUpdated) {
    resultUpdated = 0;
    printResult(collected);
  }
  if (saySomething) {
    saySomething = 0;
    printResult(NULL);
  }
  if (Serial.available() >= 3) {
    if (0xFF == Serial.read()) {
      maxOverflows = Serial.read();
      sensorCount = Serial.read();
    }
  }
}

